@@ -1,39 +1,54 @@
 /*
- * $Id: ossl_bn.c 44909 2014-02-11 14:06:51Z akr $
  * 'OpenSSL for Ruby' project
  * Copyright (C) 2001-2002  Technorama team <oss-ruby@technorama.net>
  * All rights reserved.
  */
 /*
- * This program is licenced under the same licence as Ruby.
+ * This program is licensed under the same licence as Ruby.
  * (See the file 'LICENCE'.)
  */
 /* modified by Michal Rokos <m.rokos@sh.cvut.cz> */
 #include "ossl.h"
 
-#define WrapBN(klass, obj, bn) do { \
+#define NewBN(klass) \
+  TypedData_Wrap_Struct((klass), &ossl_bn_type, 0)
+#define SetBN(obj, bn) do { \
   if (!(bn)) { \
     ossl_raise(rb_eRuntimeError, "BN wasn't initialized!"); \
   } \
-  (obj) = Data_Wrap_Struct((klass), 0, BN_clear_free, (bn)); \
+  RTYPEDDATA_DATA(obj) = (bn); \
 } while (0)
 
 #define GetBN(obj, bn) do { \
-  Data_Get_Struct((obj), BIGNUM, (bn)); \
+  TypedData_Get_Struct((obj), BIGNUM, &ossl_bn_type, (bn)); \
   if (!(bn)) { \
     ossl_raise(rb_eRuntimeError, "BN wasn't initialized!"); \
   } \
 } while (0)
 
-#define SafeGetBN(obj, bn) do { \
-  OSSL_Check_Kind((obj), cBN); \
-  GetBN((obj), (bn)); \
-} while (0)
+static void
+ossl_bn_free(void *ptr)
+{
+    BN_clear_free(ptr);
+}
+
+static const rb_data_type_t ossl_bn_type = {
+    "OpenSSL/BN",
+    {
+	0, ossl_bn_free,
+    },
+    0, 0, RUBY_TYPED_FREE_IMMEDIATELY,
+};
 
 /*
  * Classes
  */
 VALUE cBN;
+
+/* Document-class: OpenSSL::BNError
+ *
+ * Generic Error for all of OpenSSL::BN (big num)
+ */
 VALUE eBNError;
 
 /*
@@ -45,36 +60,90 @@ ossl_bn_new(const BIGNUM *bn)
     BIGNUM *newbn;
     VALUE obj;
 
+    obj = NewBN(cBN);
     newbn = bn ? BN_dup(bn) : BN_new();
     if (!newbn) {
 	ossl_raise(eBNError, NULL);
     }
-    WrapBN(cBN, obj, newbn);
+    SetBN(obj, newbn);
 
     return obj;
 }
 
-BIGNUM *
-GetBNPtr(VALUE obj)
+static BIGNUM *
+integer_to_bnptr(VALUE obj, BIGNUM *orig)
 {
-    BIGNUM *bn = NULL;
+    BIGNUM *bn;
 
-    if (RTEST(rb_obj_is_kind_of(obj, cBN))) {
-	GetBN(obj, bn);
-    } else switch (TYPE(obj)) {
-    case T_FIXNUM:
-    case T_BIGNUM:
-	obj = rb_String(obj);
-	if (!BN_dec2bn(&bn, StringValuePtr(obj))) {
-	    ossl_raise(eBNError, NULL);
+    if (FIXNUM_P(obj)) {
+	long i;
+	unsigned char bin[sizeof(long)];
+	long n = FIX2LONG(obj);
+	unsigned long un = labs(n);
+
+	for (i = sizeof(long) - 1; 0 <= i; i--) {
+	    bin[i] = un & 0xff;
+	    un >>= 8;
 	}
-	WrapBN(cBN, obj, bn); /* Handle potencial mem leaks */
-	break;
-    case T_NIL:
-	break;
-    default:
-	ossl_raise(rb_eTypeError, "Cannot convert into OpenSSL::BN");
+
+	bn = BN_bin2bn(bin, sizeof(bin), orig);
+	if (!bn)
+	    ossl_raise(eBNError, "BN_bin2bn");
+	if (n < 0)
+	    BN_set_negative(bn, 1);
     }
+    else { /* assuming Bignum */
+	size_t len = rb_absint_size(obj, NULL);
+	unsigned char *bin;
+	VALUE buf;
+	int sign;
+
+	if (INT_MAX < len) {
+	    rb_raise(eBNError, "bignum too long");
+	}
+	bin = (unsigned char*)ALLOCV_N(unsigned char, buf, len);
+	sign = rb_integer_pack(obj, bin, len, 1, 0, INTEGER_PACK_BIG_ENDIAN);
+
+	bn = BN_bin2bn(bin, (int)len, orig);
+	ALLOCV_END(buf);
+	if (!bn)
+	    ossl_raise(eBNError, "BN_bin2bn");
+	if (sign < 0)
+	    BN_set_negative(bn, 1);
+    }
+
+    return bn;
+}
+
+static VALUE
+try_convert_to_bn(VALUE obj)
+{
+    BIGNUM *bn;
+    VALUE newobj = Qnil;
+
+    if (rb_obj_is_kind_of(obj, cBN))
+	return obj;
+    if (RB_INTEGER_TYPE_P(obj)) {
+	newobj = NewBN(cBN); /* Handle potential mem leaks */
+	bn = integer_to_bnptr(obj, NULL);
+	SetBN(newobj, bn);
+    }
+
+    return newobj;
+}
+
+BIGNUM *
+ossl_bn_value_ptr(volatile VALUE *ptr)
+{
+    VALUE tmp;
+    BIGNUM *bn;
+
+    tmp = try_convert_to_bn(*ptr);
+    if (NIL_P(tmp))
+	ossl_raise(rb_eTypeError, "Cannot convert into OpenSSL::BN");
+    GetBN(tmp, bn);
+    *ptr = tmp;
+
     return bn;
 }
 
@@ -92,23 +161,25 @@ static VALUE
 ossl_bn_alloc(VALUE klass)
 {
     BIGNUM *bn;
-    VALUE obj;
+    VALUE obj = NewBN(klass);
 
     if (!(bn = BN_new())) {
 	ossl_raise(eBNError, NULL);
     }
-    WrapBN(klass, obj, bn);
+    SetBN(obj, bn);
 
     return obj;
 }
 
 /*
  * call-seq:
- *    BN.new => aBN
- *    BN.new(bn) => aBN
- *    BN.new(integer) => aBN
- *    BN.new(string) => aBN
- *    BN.new(string, 0 | 2 | 10 | 16) => aBN
+ *    OpenSSL::BN.new => aBN
+ *    OpenSSL::BN.new(bn) => aBN
+ *    OpenSSL::BN.new(integer) => aBN
+ *    OpenSSL::BN.new(string) => aBN
+ *    OpenSSL::BN.new(string, 0 | 2 | 10 | 16) => aBN
+ *
+ * Construct a new OpenSSL BIGNUM object.
  */
 static VALUE
 ossl_bn_initialize(int argc, VALUE *argv, VALUE self)
@@ -121,45 +192,13 @@ ossl_bn_initialize(int argc, VALUE *argv, VALUE self)
 	base = NUM2INT(bs);
     }
 
-    if (RB_TYPE_P(str, T_FIXNUM)) {
-	long i;
-	unsigned char bin[sizeof(long)];
-	long n = FIX2LONG(str);
-	unsigned long un = labs(n);
-
-	for (i = sizeof(long) - 1; 0 <= i; i--) {
-	    bin[i] = un&0xff;
-	    un >>= 8;
-	}
-
+    if (RB_INTEGER_TYPE_P(str)) {
 	GetBN(self, bn);
-	if (!BN_bin2bn(bin, sizeof(bin), bn)) {
-	    ossl_raise(eBNError, NULL);
-	}
-	if (n < 0) BN_set_negative(bn, 1);
+	integer_to_bnptr(str, bn);
+
 	return self;
     }
-    else if (RB_TYPE_P(str, T_BIGNUM)) {
-        size_t len = rb_absint_size(str, NULL);
-	unsigned char *bin;
-	VALUE buf;
-        int sign;
 
-        if (INT_MAX < len) {
-            rb_raise(eBNError, "bignum too long");
-        }
-        bin = (unsigned char*)ALLOCV_N(unsigned char, buf, len);
-        sign = rb_integer_pack(str, bin, len, 1, 0, INTEGER_PACK_BIG_ENDIAN);
-
-	GetBN(self, bn);
-	if (!BN_bin2bn(bin, (int)len, bn)) {
-	    ALLOCV_END(buf);
-	    ossl_raise(eBNError, NULL);
-	}
-	ALLOCV_END(buf);
-	if (sign < 0) BN_set_negative(bn, 1);
-	return self;
-    }
     if (RTEST(rb_obj_is_kind_of(str, cBN))) {
 	BIGNUM *other;
 
@@ -171,26 +210,25 @@ ossl_bn_initialize(int argc, VALUE *argv, VALUE self)
 	return self;
     }
 
-    StringValue(str);
     GetBN(self, bn);
     switch (base) {
     case 0:
-	if (!BN_mpi2bn((unsigned char *)RSTRING_PTR(str), RSTRING_LENINT(str), bn)) {
+	if (!BN_mpi2bn((unsigned char *)StringValuePtr(str), RSTRING_LENINT(str), bn)) {
 	    ossl_raise(eBNError, NULL);
 	}
 	break;
     case 2:
-	if (!BN_bin2bn((unsigned char *)RSTRING_PTR(str), RSTRING_LENINT(str), bn)) {
+	if (!BN_bin2bn((unsigned char *)StringValuePtr(str), RSTRING_LENINT(str), bn)) {
 	    ossl_raise(eBNError, NULL);
 	}
 	break;
     case 10:
-	if (!BN_dec2bn(&bn, RSTRING_PTR(str))) {
+	if (!BN_dec2bn(&bn, StringValueCStr(str))) {
 	    ossl_raise(eBNError, NULL);
 	}
 	break;
     case 16:
-	if (!BN_hex2bn(&bn, RSTRING_PTR(str))) {
+	if (!BN_hex2bn(&bn, StringValueCStr(str))) {
 	    ossl_raise(eBNError, NULL);
 	}
 	break;
@@ -206,12 +244,12 @@ ossl_bn_initialize(int argc, VALUE *argv, VALUE self)
  *    bn.to_s(base) => string
  *
  * === Parameters
- * * +base+ - integer
- * * * Valid values:
- * * * * 0 - MPI
- * * * * 2 - binary
- * * * * 10 - the default
- * * * * 16 - hex
+ * * _base_ - Integer
+ *   Valid values:
+ *   * 0 - MPI
+ *   * 2 - binary
+ *   * 10 - the default
+ *   * 16 - hex
  */
 static VALUE
 ossl_bn_to_s(int argc, VALUE *argv, VALUE self)
@@ -301,11 +339,6 @@ ossl_bn_coerce(VALUE self, VALUE other)
 }
 
 #define BIGNUM_BOOL1(func)				\
-    /*							\
-     * call-seq:					\
-     *   bn.##func -> true | false			\
-     *							\
-     */							\
     static VALUE					\
     ossl_bn_##func(VALUE self)				\
     {							\
@@ -316,22 +349,51 @@ ossl_bn_coerce(VALUE self, VALUE other)
 	}						\
 	return Qfalse;					\
     }
+
+/*
+ * Document-method: OpenSSL::BN#zero?
+ * call-seq:
+ *   bn.zero? => true | false
+ */
 BIGNUM_BOOL1(is_zero)
+
+/*
+ * Document-method: OpenSSL::BN#one?
+ * call-seq:
+ *   bn.one? => true | false
+ */
 BIGNUM_BOOL1(is_one)
+
+/*
+ * Document-method: OpenSSL::BN#odd?
+ * call-seq:
+ *   bn.odd? => true | false
+ */
 BIGNUM_BOOL1(is_odd)
 
+/*
+ * call-seq:
+ *   bn.negative? => true | false
+ */
+static VALUE
+ossl_bn_is_negative(VALUE self)
+{
+    BIGNUM *bn;
+
+    GetBN(self, bn);
+    if (BN_is_zero(bn))
+	return Qfalse;
+    return BN_is_negative(bn) ? Qtrue : Qfalse;
+}
+
 #define BIGNUM_1c(func)					\
-    /*							\
-     * call-seq:					\
-     *   bn.##func -> aBN				\
-     *							\
-     */							\
     static VALUE					\
     ossl_bn_##func(VALUE self)				\
     {							\
 	BIGNUM *bn, *result;				\
 	VALUE obj;					\
 	GetBN(self, bn);				\
+	obj = NewBN(rb_obj_class(self));		\
 	if (!(result = BN_new())) {			\
 	    ossl_raise(eBNError, NULL);			\
 	}						\
@@ -339,23 +401,25 @@ BIGNUM_BOOL1(is_odd)
 	    BN_free(result);				\
 	    ossl_raise(eBNError, NULL);			\
 	}						\
-	WrapBN(CLASS_OF(self), obj, result);		\
+	SetBN(obj, result);				\
 	return obj;					\
     }
+
+/*
+ * Document-method: OpenSSL::BN#sqr
+ * call-seq:
+ *   bn.sqr => aBN
+ */
 BIGNUM_1c(sqr)
 
 #define BIGNUM_2(func)					\
-    /*							\
-     * call-seq:					\
-     *   bn.##func(bn2) -> aBN				\
-     *							\
-     */							\
     static VALUE					\
     ossl_bn_##func(VALUE self, VALUE other)		\
     {							\
 	BIGNUM *bn1, *bn2 = GetBNPtr(other), *result;	\
 	VALUE obj;					\
 	GetBN(self, bn1);				\
+	obj = NewBN(rb_obj_class(self));		\
 	if (!(result = BN_new())) {			\
 	    ossl_raise(eBNError, NULL);			\
 	}						\
@@ -363,24 +427,32 @@ BIGNUM_1c(sqr)
 	    BN_free(result);				\
 	    ossl_raise(eBNError, NULL);			\
 	}						\
-	WrapBN(CLASS_OF(self), obj, result);		\
+	SetBN(obj, result);				\
 	return obj;					\
     }
+
+/*
+ * Document-method: OpenSSL::BN#+
+ * call-seq:
+ *   bn + bn2 => aBN
+ */
 BIGNUM_2(add)
+
+/*
+ * Document-method: OpenSSL::BN#-
+ * call-seq:
+ *   bn - bn2 => aBN
+ */
 BIGNUM_2(sub)
 
 #define BIGNUM_2c(func)						\
-    /*								\
-     * call-seq:						\
-     *   bn.##func(bn2) -> aBN					\
-     *								\
-     */								\
     static VALUE						\
     ossl_bn_##func(VALUE self, VALUE other)			\
     {								\
 	BIGNUM *bn1, *bn2 = GetBNPtr(other), *result;		\
 	VALUE obj;						\
 	GetBN(self, bn1);					\
+	obj = NewBN(rb_obj_class(self));			\
 	if (!(result = BN_new())) {				\
 	    ossl_raise(eBNError, NULL);				\
 	}							\
@@ -388,28 +460,69 @@ BIGNUM_2(sub)
 	    BN_free(result);					\
 	    ossl_raise(eBNError, NULL);				\
 	}							\
-	WrapBN(CLASS_OF(self), obj, result);			\
+	SetBN(obj, result);					\
 	return obj;						\
     }
+
+/*
+ * Document-method: OpenSSL::BN#*
+ * call-seq:
+ *   bn * bn2 => aBN
+ */
 BIGNUM_2c(mul)
+
+/*
+ * Document-method: OpenSSL::BN#%
+ * call-seq:
+ *   bn % bn2 => aBN
+ */
 BIGNUM_2c(mod)
+
+/*
+ * Document-method: OpenSSL::BN#**
+ * call-seq:
+ *   bn ** bn2 => aBN
+ */
 BIGNUM_2c(exp)
+
+/*
+ * Document-method: OpenSSL::BN#gcd
+ * call-seq:
+ *   bn.gcd(bn2) => aBN
+ */
 BIGNUM_2c(gcd)
+
+/*
+ * Document-method: OpenSSL::BN#mod_sqr
+ * call-seq:
+ *   bn.mod_sqr(bn2) => aBN
+ */
 BIGNUM_2c(mod_sqr)
+
+/*
+ * Document-method: OpenSSL::BN#mod_inverse
+ * call-seq:
+ *   bn.mod_inverse(bn2) => aBN
+ */
 BIGNUM_2c(mod_inverse)
 
 /*
  * call-seq:
  *    bn1 / bn2 => [result, remainder]
+ *
+ * Division of OpenSSL::BN instances
  */
 static VALUE
 ossl_bn_div(VALUE self, VALUE other)
 {
     BIGNUM *bn1, *bn2 = GetBNPtr(other), *r1, *r2;
-    VALUE obj1, obj2;
+    VALUE klass, obj1, obj2;
 
     GetBN(self, bn1);
 
+    klass = rb_obj_class(self);
+    obj1 = NewBN(klass);
+    obj2 = NewBN(klass);
     if (!(r1 = BN_new())) {
 	ossl_raise(eBNError, NULL);
     }
@@ -422,18 +535,13 @@ ossl_bn_div(VALUE self, VALUE other)
 	BN_free(r2);
 	ossl_raise(eBNError, NULL);
     }
-    WrapBN(CLASS_OF(self), obj1, r1);
-    WrapBN(CLASS_OF(self), obj2, r2);
+    SetBN(obj1, r1);
+    SetBN(obj2, r2);
 
     return rb_ary_new3(2, obj1, obj2);
 }
 
 #define BIGNUM_3c(func)						\
-    /*								\
-     * call-seq:						\
-     *   bn.##func(bn1, bn2) -> aBN				\
-     *								\
-     */								\
     static VALUE						\
     ossl_bn_##func(VALUE self, VALUE other1, VALUE other2)	\
     {								\
@@ -441,6 +549,7 @@ ossl_bn_div(VALUE self, VALUE other)
 	BIGNUM *bn3 = GetBNPtr(other2), *result;		\
 	VALUE obj;						\
 	GetBN(self, bn1);					\
+	obj = NewBN(rb_obj_class(self));			\
 	if (!(result = BN_new())) {				\
 	    ossl_raise(eBNError, NULL);				\
 	}							\
@@ -448,20 +557,39 @@ ossl_bn_div(VALUE self, VALUE other)
 	    BN_free(result);					\
 	    ossl_raise(eBNError, NULL);				\
 	}							\
-	WrapBN(CLASS_OF(self), obj, result);			\
+	SetBN(obj, result);					\
 	return obj;						\
     }
+
+/*
+ * Document-method: OpenSSL::BN#mod_add
+ * call-seq:
+ *   bn.mod_add(bn1, bn2) -> aBN
+ */
 BIGNUM_3c(mod_add)
+
+/*
+ * Document-method: OpenSSL::BN#mod_sub
+ * call-seq:
+ *   bn.mod_sub(bn1, bn2) -> aBN
+ */
 BIGNUM_3c(mod_sub)
+
+/*
+ * Document-method: OpenSSL::BN#mod_mul
+ * call-seq:
+ *   bn.mod_mul(bn1, bn2) -> aBN
+ */
 BIGNUM_3c(mod_mul)
+
+/*
+ * Document-method: OpenSSL::BN#mod_exp
+ * call-seq:
+ *   bn.mod_exp(bn1, bn2) -> aBN
+ */
 BIGNUM_3c(mod_exp)
 
 #define BIGNUM_BIT(func)				\
-    /*							\
-     * call-seq:					\
-     *   bn.##func(bit) -> self				\
-     *							\
-     */							\
     static VALUE					\
     ossl_bn_##func(VALUE self, VALUE bit)		\
     {							\
@@ -472,13 +600,33 @@ BIGNUM_3c(mod_exp)
 	}						\
 	return self;					\
     }
+
+/*
+ * Document-method: OpenSSL::BN#set_bit!
+ * call-seq:
+ *   bn.set_bit!(bit) -> self
+ */
 BIGNUM_BIT(set_bit)
+
+/*
+ * Document-method: OpenSSL::BN#clear_bit!
+ * call-seq:
+ *   bn.clear_bit!(bit) -> self
+ */
 BIGNUM_BIT(clear_bit)
+
+/*
+ * Document-method: OpenSSL::BN#mask_bit!
+ * call-seq:
+ *   bn.mask_bit!(bit) -> self
+ */
 BIGNUM_BIT(mask_bits)
 
 /*
  * call-seq:
- *    bn.bit_set?(bit) => true | false
+ *   bn.bit_set?(bit) => true | false
+ *
+ * Tests bit _bit_ in _bn_ and returns +true+ if set, +false+ if not set.
  */
 static VALUE
 ossl_bn_is_bit_set(VALUE self, VALUE bit)
@@ -495,11 +643,6 @@ ossl_bn_is_bit_set(VALUE self, VALUE bit)
 }
 
 #define BIGNUM_SHIFT(func)				\
-    /*							\
-     * call-seq:					\
-     *   bn.##func(bits) -> aBN				\
-     *							\
-     */							\
     static VALUE					\
     ossl_bn_##func(VALUE self, VALUE bits)		\
     {							\
@@ -508,6 +651,7 @@ ossl_bn_is_bit_set(VALUE self, VALUE bit)
 	VALUE obj;					\
 	b = NUM2INT(bits);				\
 	GetBN(self, bn);				\
+	obj = NewBN(rb_obj_class(self));		\
 	if (!(result = BN_new())) {			\
 		ossl_raise(eBNError, NULL);		\
 	}						\
@@ -515,18 +659,25 @@ ossl_bn_is_bit_set(VALUE self, VALUE bit)
 		BN_free(result);			\
 		ossl_raise(eBNError, NULL);		\
 	}						\
-	WrapBN(CLASS_OF(self), obj, result);		\
+	SetBN(obj, result);				\
 	return obj;					\
     }
+
+/*
+ * Document-method: OpenSSL::BN#<<
+ * call-seq:
+ *   bn << bits -> aBN
+ */
 BIGNUM_SHIFT(lshift)
+
+/*
+ * Document-method: OpenSSL::BN#>>
+ * call-seq:
+ *   bn >> bits -> aBN
+ */
 BIGNUM_SHIFT(rshift)
 
 #define BIGNUM_SELF_SHIFT(func)				\
-    /*							\
-     * call-seq:					\
-     *   bn.##func!(bits) -> self			\
-     *							\
-     */							\
     static VALUE					\
     ossl_bn_self_##func(VALUE self, VALUE bits)		\
     {							\
@@ -538,15 +689,22 @@ BIGNUM_SHIFT(rshift)
 		ossl_raise(eBNError, NULL);		\
 	return self;					\
     }
+
+/*
+ * Document-method: OpenSSL::BN#lshift!
+ * call-seq:
+ *   bn.lshift!(bits) -> self
+ */
 BIGNUM_SELF_SHIFT(lshift)
+
+/*
+ * Document-method: OpenSSL::BN#rshift!
+ * call-seq:
+ *   bn.rshift!(bits) -> self
+ */
 BIGNUM_SELF_SHIFT(rshift)
 
 #define BIGNUM_RAND(func)					\
-    /*								\
-     * call-seq:						\
-     *   BN.##func(bits [, fill [, odd]]) -> aBN		\
-     *								\
-     */								\
     static VALUE						\
     ossl_bn_s_##func(int argc, VALUE *argv, VALUE klass)	\
     {								\
@@ -562,6 +720,7 @@ BIGNUM_SELF_SHIFT(rshift)
 	    top = NUM2INT(fill);				\
 	}							\
 	b = NUM2INT(bits);					\
+	obj = NewBN(klass);					\
 	if (!(result = BN_new())) {				\
 	    ossl_raise(eBNError, NULL);				\
 	}							\
@@ -569,23 +728,28 @@ BIGNUM_SELF_SHIFT(rshift)
 	    BN_free(result);					\
 	    ossl_raise(eBNError, NULL);				\
 	}							\
-	WrapBN(klass, obj, result);				\
+	SetBN(obj, result);					\
 	return obj;						\
     }
+
+/*
+ * Document-method: OpenSSL::BN.rand
+ *   BN.rand(bits [, fill [, odd]]) -> aBN
+ */
 BIGNUM_RAND(rand)
+
+/*
+ * Document-method: OpenSSL::BN.pseudo_rand
+ *   BN.pseudo_rand(bits [, fill [, odd]]) -> aBN
+ */
 BIGNUM_RAND(pseudo_rand)
 
 #define BIGNUM_RAND_RANGE(func)					\
-    /*								\
-     * call-seq:						\
-     *   BN.##func(range) -> aBN				\
-     *								\
-     */								\
     static VALUE						\
     ossl_bn_s_##func##_range(VALUE klass, VALUE range)		\
     {								\
 	BIGNUM *bn = GetBNPtr(range), *result;			\
-	VALUE obj;						\
+	VALUE obj = NewBN(klass);				\
 	if (!(result = BN_new())) {				\
 	    ossl_raise(eBNError, NULL);				\
 	}							\
@@ -593,21 +757,39 @@ BIGNUM_RAND(pseudo_rand)
 	    BN_free(result);					\
 	    ossl_raise(eBNError, NULL);				\
 	}							\
-	WrapBN(klass, obj, result);				\
+	SetBN(obj, result);					\
 	return obj;						\
     }
+
+/*
+ * Document-method: OpenSSL::BN.rand_range
+ * call-seq:
+ *   BN.rand_range(range) -> aBN
+ *
+ */
 BIGNUM_RAND_RANGE(rand)
+
+/*
+ * Document-method: OpenSSL::BN.pseudo_rand_range
+ * call-seq:
+ *   BN.pseudo_rand_range(range) -> aBN
+ *
+ */
 BIGNUM_RAND_RANGE(pseudo_rand)
 
 /*
  * call-seq:
  *    BN.generate_prime(bits, [, safe [, add [, rem]]]) => bn
  *
+ * Generates a random prime number of bit length _bits_. If _safe_ is set to
+ * +true+, generates a safe prime. If _add_ is specified, generates a prime that
+ * fulfills condition <tt>p % add = rem</tt>.
+ *
  * === Parameters
- * * +bits+ - integer
- * * +safe+ - boolean
- * * +add+ - BN
- * * +rem+ - BN
+ * * _bits_ - integer
+ * * _safe_ - boolean
+ * * _add_ - BN
+ * * _rem_ - BN
  */
 static VALUE
 ossl_bn_s_generate_prime(int argc, VALUE *argv, VALUE klass)
@@ -627,32 +809,40 @@ ossl_bn_s_generate_prime(int argc, VALUE *argv, VALUE klass)
 	add = GetBNPtr(vadd);
 	rem = NIL_P(vrem) ? NULL : GetBNPtr(vrem);
     }
+    obj = NewBN(klass);
     if (!(result = BN_new())) {
 	ossl_raise(eBNError, NULL);
     }
-    if (!BN_generate_prime(result, num, safe, add, rem, NULL, NULL)) {
+    if (!BN_generate_prime_ex(result, num, safe, add, rem, NULL)) {
 	BN_free(result);
 	ossl_raise(eBNError, NULL);
     }
-    WrapBN(klass, obj, result);
+    SetBN(obj, result);
 
     return obj;
 }
 
 #define BIGNUM_NUM(func)			\
-    /*							\
-     * call-seq:					\
-     *   bn.##func -> integer				\
-     *							\
-     */							\
     static VALUE 				\
     ossl_bn_##func(VALUE self)			\
     {						\
 	BIGNUM *bn;				\
 	GetBN(self, bn);			\
-	return INT2FIX(BN_##func(bn));		\
+	return INT2NUM(BN_##func(bn));		\
     }
+
+/*
+ * Document-method: OpenSSL::BN#num_bytes
+ * call-seq:
+ *   bn.num_bytes => integer
+ */
 BIGNUM_NUM(num_bytes)
+
+/*
+ * Document-method: OpenSSL::BN#num_bits
+ * call-seq:
+ *   bn.num_bits => integer
+ */
 BIGNUM_NUM(num_bits)
 
 static VALUE
@@ -673,38 +863,151 @@ ossl_bn_copy(VALUE self, VALUE other)
     return self;
 }
 
+/*
+ * call-seq:
+ *   +bn -> aBN
+ */
+static VALUE
+ossl_bn_uplus(VALUE self)
+{
+    return self;
+}
+
+/*
+ * call-seq:
+ *   -bn -> aBN
+ */
+static VALUE
+ossl_bn_uminus(VALUE self)
+{
+    VALUE obj;
+    BIGNUM *bn1, *bn2;
+
+    GetBN(self, bn1);
+    obj = NewBN(cBN);
+    bn2 = BN_dup(bn1);
+    if (!bn2)
+	ossl_raise(eBNError, "BN_dup");
+    SetBN(obj, bn2);
+    BN_set_negative(bn2, !BN_is_negative(bn2));
+
+    return obj;
+}
+
 #define BIGNUM_CMP(func)				\
-    /*							\
-     * call-seq:					\
-     *   bn.##func(bn2) -> integer			\
-     *							\
-     */							\
     static VALUE					\
     ossl_bn_##func(VALUE self, VALUE other)		\
     {							\
 	BIGNUM *bn1, *bn2 = GetBNPtr(other);		\
 	GetBN(self, bn1);				\
-	return INT2FIX(BN_##func(bn1, bn2));		\
+	return INT2NUM(BN_##func(bn1, bn2));		\
     }
+
+/*
+ * Document-method: OpenSSL::BN#cmp
+ * call-seq:
+ *   bn.cmp(bn2) => integer
+ */
+/*
+ * Document-method: OpenSSL::BN#<=>
+ * call-seq:
+ *   bn <=> bn2 => integer
+ */
 BIGNUM_CMP(cmp)
+
+/*
+ * Document-method: OpenSSL::BN#ucmp
+ * call-seq:
+ *   bn.ucmp(bn2) => integer
+ */
 BIGNUM_CMP(ucmp)
 
+/*
+ *  call-seq:
+ *     bn == obj => true or false
+ *
+ *  Returns +true+ only if _obj_ has the same value as _bn_. Contrast this
+ *  with OpenSSL::BN#eql?, which requires obj to be OpenSSL::BN.
+ */
 static VALUE
-ossl_bn_eql(VALUE self, VALUE other)
+ossl_bn_eq(VALUE self, VALUE other)
 {
-    if (ossl_bn_cmp(self, other) == INT2FIX(0)) {
+    BIGNUM *bn1, *bn2;
+
+    GetBN(self, bn1);
+    other = try_convert_to_bn(other);
+    if (NIL_P(other))
+	return Qfalse;
+    GetBN(other, bn2);
+
+    if (!BN_cmp(bn1, bn2)) {
 	return Qtrue;
     }
     return Qfalse;
 }
 
+/*
+ *  call-seq:
+ *     bn.eql?(obj) => true or false
+ *
+ *  Returns <code>true</code> only if <i>obj</i> is a
+ *  <code>OpenSSL::BN</code> with the same value as <i>bn</i>. Contrast this
+ *  with OpenSSL::BN#==, which performs type conversions.
+ */
+static VALUE
+ossl_bn_eql(VALUE self, VALUE other)
+{
+    BIGNUM *bn1, *bn2;
+
+    if (!rb_obj_is_kind_of(other, cBN))
+	return Qfalse;
+    GetBN(self, bn1);
+    GetBN(other, bn2);
+
+    return BN_cmp(bn1, bn2) ? Qfalse : Qtrue;
+}
+
+/*
+ *  call-seq:
+ *     bn.hash => Integer
+ *
+ *  Returns a hash code for this object.
+ *
+ *  See also Object#hash.
+ */
+static VALUE
+ossl_bn_hash(VALUE self)
+{
+    BIGNUM *bn;
+    VALUE hash;
+    unsigned char *buf;
+    int len;
+
+    GetBN(self, bn);
+    len = BN_num_bytes(bn);
+    buf = xmalloc(len);
+    if (BN_bn2bin(bn, buf) != len) {
+	xfree(buf);
+	ossl_raise(eBNError, NULL);
+    }
+
+    hash = ST2FIX(rb_memhash(buf, len));
+    xfree(buf);
+
+    return hash;
+}
+
 /*
  * call-seq:
  *    bn.prime? => true | false
  *    bn.prime?(checks) => true | false
  *
+ * Performs a Miller-Rabin probabilistic primality test with _checks_
+ * iterations. If _checks_ is not specified, a number of iterations is used
+ * that yields a false positive rate of at most 2^-80 for random input.
+ *
  * === Parameters
- * * +checks+ - integer
+ * * _checks_ - integer
  */
 static VALUE
 ossl_bn_is_prime(int argc, VALUE *argv, VALUE self)
@@ -717,7 +1020,7 @@ ossl_bn_is_prime(int argc, VALUE *argv, VALUE self)
 	checks = NUM2INT(vchecks);
     }
     GetBN(self, bn);
-    switch (BN_is_prime(bn, checks, NULL, ossl_bn_ctx, NULL)) {
+    switch (BN_is_prime_ex(bn, checks, ossl_bn_ctx, NULL)) {
     case 1:
 	return Qtrue;
     case 0:
@@ -735,9 +1038,12 @@ ossl_bn_is_prime(int argc, VALUE *argv, VALUE self)
  *    bn.prime_fasttest?(checks) => true | false
  *    bn.prime_fasttest?(checks, trial_div) => true | false
  *
+ * Performs a Miller-Rabin primality test. This is same as #prime? except this
+ * first attempts trial divisions with some small primes.
+ *
  * === Parameters
- * * +checks+ - integer
- * * +trial_div+ - boolean
+ * * _checks_ - integer
+ * * _trial_div_ - boolean
  */
 static VALUE
 ossl_bn_is_prime_fasttest(int argc, VALUE *argv, VALUE self)
@@ -756,7 +1062,7 @@ ossl_bn_is_prime_fasttest(int argc, VALUE *argv, VALUE self)
     if (vtrivdiv == Qfalse) {
 	do_trial_division = 0;
     }
-    switch (BN_is_prime_fasttest(bn, checks, NULL, ossl_bn_ctx, NULL, do_trial_division)) {
+    switch (BN_is_prime_fasttest_ex(bn, checks, ossl_bn_ctx, do_trial_division, NULL)) {
     case 1:
 	return Qtrue;
     case 0:
@@ -773,10 +1079,11 @@ ossl_bn_is_prime_fasttest(int argc, VALUE *argv, VALUE self)
  * (NOTE: ordering of methods is the same as in 'man bn')
  */
 void
-Init_ossl_bn()
+Init_ossl_bn(void)
 {
 #if 0
-    mOSSL = rb_define_module("OpenSSL"); /* let rdoc know about mOSSL */
+    mOSSL = rb_define_module("OpenSSL");
+    eOSSLError = rb_define_class_under(mOSSL, "OpenSSLError", rb_eStandardError);
 #endif
 
     if (!(ossl_bn_ctx = BN_CTX_new())) {
@@ -790,7 +1097,7 @@ Init_ossl_bn()
     rb_define_alloc_func(cBN, ossl_bn_alloc);
     rb_define_method(cBN, "initialize", ossl_bn_initialize, -1);
 
-    rb_define_copy_func(cBN, ossl_bn_copy);
+    rb_define_method(cBN, "initialize_copy", ossl_bn_copy, 1);
     rb_define_method(cBN, "copy", ossl_bn_copy, 1);
 
     /* swap (=coerce?) */
@@ -799,6 +1106,9 @@ Init_ossl_bn()
     rb_define_method(cBN, "num_bits", ossl_bn_num_bits, 0);
     /* num_bits_word */
 
+    rb_define_method(cBN, "+@", ossl_bn_uplus, 0);
+    rb_define_method(cBN, "-@", ossl_bn_uminus, 0);
+
     rb_define_method(cBN, "+", ossl_bn_add, 1);
     rb_define_method(cBN, "-", ossl_bn_sub, 1);
     rb_define_method(cBN, "*", ossl_bn_mul, 1);
@@ -825,12 +1135,14 @@ Init_ossl_bn()
     rb_define_alias(cBN, "<=>", "cmp");
     rb_define_method(cBN, "ucmp", ossl_bn_ucmp, 1);
     rb_define_method(cBN, "eql?", ossl_bn_eql, 1);
-    rb_define_alias(cBN, "==", "eql?");
-    rb_define_alias(cBN, "===", "eql?");
+    rb_define_method(cBN, "hash", ossl_bn_hash, 0);
+    rb_define_method(cBN, "==", ossl_bn_eq, 1);
+    rb_define_alias(cBN, "===", "==");
     rb_define_method(cBN, "zero?", ossl_bn_is_zero, 0);
     rb_define_method(cBN, "one?", ossl_bn_is_one, 0);
     /* is_word */
     rb_define_method(cBN, "odd?", ossl_bn_is_odd, 0);
+    rb_define_method(cBN, "negative?", ossl_bn_is_negative, 0);
 
     /* zero
      * one
@@ -845,6 +1157,7 @@ Init_ossl_bn()
 
     rb_define_singleton_method(cBN, "generate_prime", ossl_bn_s_generate_prime, -1);
     rb_define_method(cBN, "prime?", ossl_bn_is_prime, -1);
+    rb_define_method(cBN, "prime_fasttest?", ossl_bn_is_prime_fasttest, -1);
 
     rb_define_method(cBN, "set_bit!", ossl_bn_set_bit, 1);
     rb_define_method(cBN, "clear_bit!", ossl_bn_clear_bit, 1);
@@ -886,11 +1199,4 @@ Init_ossl_bn()
 
     /* RECiProcal
      * MONTgomery */
-
-    /*
-     * TODO:
-     * Where to belong these?
-     */
-    rb_define_method(cBN, "prime_fasttest?", ossl_bn_is_prime_fasttest, -1);
 }
-
