@@ -85,8 +85,8 @@ rb_construct_expanded_load_path(enum expand_type type, int *has_relative, int *h
 	if (is_string)
 	    rb_str_freeze(path);
 	as_str = rb_get_path_check_convert(path, as_str, level);
-	expanded_path = rb_file_expand_path_fast(as_str, Qnil);
-	rb_str_freeze(expanded_path);
+	expanded_path = rb_check_realpath(Qnil, as_str);
+	if (NIL_P(expanded_path)) expanded_path = as_str;
 	rb_ary_push(ary, rb_fstring(expanded_path));
     }
     rb_obj_freeze(ary);
@@ -230,10 +230,10 @@ features_index_add(VALUE feature, VALUE offset)
     feature_end = feature_str + RSTRING_LEN(feature);
 
     for (ext = feature_end; ext > feature_str; ext--)
-      if (*ext == '.' || *ext == '/')
-	break;
+	if (*ext == '.' || *ext == '/')
+	    break;
     if (*ext != '.')
-      ext = NULL;
+	ext = NULL;
     /* Now `ext` points to the only string matching %r{^\.[^./]*$} that is
        at the end of `feature`, or is NULL if there is no such string. */
 
@@ -571,23 +571,23 @@ rb_provide_feature(VALUE feature)
 void
 rb_provide(const char *feature)
 {
-    rb_provide_feature(rb_usascii_str_new2(feature));
+    rb_provide_feature(rb_fstring_cstr(feature));
 }
 
 NORETURN(static void load_failed(VALUE));
+const rb_iseq_t *rb_iseq_load_iseq(VALUE fname);
 
 static int
 rb_load_internal0(rb_thread_t *th, VALUE fname, int wrap)
 {
-    int state;
+    enum ruby_tag_type state;
     volatile VALUE wrapper = th->top_wrapper;
     volatile VALUE self = th->top_self;
-    volatile int mild_compile_error;
 #if !defined __GNUC__
     rb_thread_t *volatile th0 = th;
 #endif
 
-    th->errinfo = Qnil; /* ensure */
+    th->ec.errinfo = Qnil; /* ensure */
 
     if (!wrap) {
 	th->top_wrapper = 0;
@@ -599,17 +599,22 @@ rb_load_internal0(rb_thread_t *th, VALUE fname, int wrap)
 	rb_extend_object(th->top_self, th->top_wrapper);
     }
 
-    mild_compile_error = th->mild_compile_error;
     TH_PUSH_TAG(th);
     state = EXEC_TAG();
-    if (state == 0) {
+    if (state == TAG_NONE) {
 	NODE *node;
-	rb_iseq_t *iseq;
+	const rb_iseq_t *iseq;
 
-	th->mild_compile_error++;
-	node = (NODE *)rb_load_file_str(fname);
-	iseq = rb_iseq_new_top(node, rb_str_new2("<top (required)>"), fname, rb_realpath_internal(Qnil, fname, 1), NULL);
-	th->mild_compile_error--;
+	if ((iseq = rb_iseq_load_iseq(fname)) != NULL) {
+	    /* OK */
+	}
+	else {
+	    VALUE parser = rb_parser_new();
+	    rb_parser_set_context(parser, NULL, FALSE);
+	    node = (NODE *)rb_parser_load_file(parser, fname);
+	    iseq = rb_iseq_new_top(node, rb_fstring_cstr("<top (required)>"),
+			    fname, rb_realpath_internal(Qnil, fname, 1), NULL);
+	}
 	rb_iseq_eval(iseq);
     }
     TH_POP_TAG();
@@ -618,18 +623,19 @@ rb_load_internal0(rb_thread_t *th, VALUE fname, int wrap)
     th = th0;
     fname = RB_GC_GUARD(fname);
 #endif
-    th->mild_compile_error = mild_compile_error;
     th->top_self = self;
     th->top_wrapper = wrapper;
 
     if (state) {
+	/* usually state == TAG_RAISE only, except for
+	 * rb_iseq_load_iseq case */
 	VALUE exc = rb_vm_make_jump_tag_but_local_jump(state, Qundef);
 	if (NIL_P(exc)) return state;
-	th->errinfo = exc;
+	th->ec.errinfo = exc;
 	return TAG_RAISE;
     }
 
-    if (!NIL_P(th->errinfo)) {
+    if (!NIL_P(th->ec.errinfo)) {
 	/* exception during load */
 	return TAG_RAISE;
     }
@@ -642,8 +648,8 @@ rb_load_internal(VALUE fname, int wrap)
     rb_thread_t *curr_th = GET_THREAD();
     int state = rb_load_internal0(curr_th, fname, wrap);
     if (state) {
-	if (state == TAG_RAISE) rb_exc_raise(curr_th->errinfo);
-	JUMP_TAG(state);
+	if (state == TAG_RAISE) rb_exc_raise(curr_th->ec.errinfo);
+	TH_JUMP_TAG(curr_th, state);
     }
 }
 
@@ -662,19 +668,19 @@ rb_load(VALUE fname, int wrap)
 }
 
 void
-rb_load_protect(VALUE fname, int wrap, int *state)
+rb_load_protect(VALUE fname, int wrap, int *pstate)
 {
-    int status;
+    enum ruby_tag_type state;
     volatile VALUE path = 0;
 
     PUSH_TAG();
-    if ((status = EXEC_TAG()) == 0) {
+    if ((state = EXEC_TAG()) == TAG_NONE) {
 	path = file_to_load(fname);
     }
     POP_TAG();
-    if (!status) status = rb_load_internal0(GET_THREAD(), path, wrap);
-    if (state)
-	*state = status;
+
+    if (state == TAG_NONE) state = rb_load_internal0(GET_THREAD(), path, wrap);
+    if (state != TAG_NONE) *pstate = state;
 }
 
 /*
@@ -698,10 +704,10 @@ rb_f_load(int argc, VALUE *argv)
 
     rb_scan_args(argc, argv, "11", &fname, &wrap);
 
-    RUBY_DTRACE_HOOK(LOAD_ENTRY, StringValuePtr(fname));
-
-    orig_fname = FilePathValue(fname);
+    orig_fname = rb_get_path_check_to_string(fname, rb_safe_level());
     fname = rb_str_encode_ospath(orig_fname);
+    RUBY_DTRACE_HOOK(LOAD_ENTRY, RSTRING_PTR(orig_fname));
+
     path = rb_find_file(fname);
     if (!path) {
 	if (!rb_file_load_ok(RSTRING_PTR(fname)))
@@ -710,11 +716,13 @@ rb_f_load(int argc, VALUE *argv)
     }
     rb_load_internal(path, RTEST(wrap));
 
-    RUBY_DTRACE_HOOK(LOAD_RETURN, StringValuePtr(fname));
+    RUBY_DTRACE_HOOK(LOAD_RETURN, RSTRING_PTR(orig_fname));
 
     return Qtrue;
 }
 
+extern VALUE rb_mWarning;
+
 static char *
 load_lock(const char *ftptr)
 {
@@ -728,7 +736,7 @@ load_lock(const char *ftptr)
 	st_insert(loading_tbl, (st_data_t)ftptr, data);
 	return (char *)ftptr;
     }
-    else if (RB_TYPE_P((VALUE)data, T_IMEMO) && imemo_type((VALUE)data) == imemo_memo) {
+    else if (imemo_type_p(data, imemo_memo)) {
 	struct MEMO *memo = MEMO_CAST(data);
 	void (*init)(void) = (void (*)(void))memo->u3.func;
 	data = (st_data_t)rb_thread_shield_new();
@@ -737,14 +745,12 @@ load_lock(const char *ftptr)
 	return (char *)"";
     }
     if (RTEST(ruby_verbose)) {
-	rb_warning("loading in progress, circular require considered harmful - %s", ftptr);
-	rb_backtrace_print_to(rb_stderr);
+	VALUE warning = rb_warning_string("loading in progress, circular require considered harmful - %s", ftptr);
+	rb_backtrace_each(rb_str_append, warning);
+	rb_warning_warn(rb_mWarning, warning);
     }
     switch (rb_thread_shield_wait((VALUE)data)) {
       case Qfalse:
-	data = (st_data_t)ftptr;
-	st_insert(loading_tbl, data, (st_data_t)rb_thread_shield_new());
-	return 0;
       case Qnil:
 	return 0;
     }
@@ -756,7 +762,12 @@ release_thread_shield(st_data_t *key, st_data_t *value, st_data_t done, int exis
 {
     VALUE thread_shield = (VALUE)*value;
     if (!existing) return ST_STOP;
-    if (done ? rb_thread_shield_destroy(thread_shield) : rb_thread_shield_release(thread_shield)) {
+    if (done) {
+	rb_thread_shield_destroy(thread_shield);
+	/* Delete the entry even if there are waiting threads, because they
+	 * won't load the file and won't delete the entry. */
+    }
+    else if (rb_thread_shield_release(thread_shield)) {
 	/* still in-use */
 	return ST_CONTINUE;
     }
@@ -952,32 +963,30 @@ rb_require_internal(VALUE fname, int safe)
 {
     volatile int result = -1;
     rb_thread_t *th = GET_THREAD();
-    volatile VALUE errinfo = th->errinfo;
-    int state;
+    volatile VALUE errinfo = th->ec.errinfo;
+    enum ruby_tag_type state;
     struct {
 	int safe;
     } volatile saved;
     char *volatile ftptr = 0;
+    VALUE path;
 
-    RUBY_DTRACE_HOOK(REQUIRE_ENTRY, StringValuePtr(fname));
+    fname = rb_get_path_check(fname, safe);
+    path = rb_str_encode_ospath(fname);
+    RUBY_DTRACE_HOOK(REQUIRE_ENTRY, RSTRING_PTR(fname));
 
     TH_PUSH_TAG(th);
     saved.safe = rb_safe_level();
-    if ((state = EXEC_TAG()) == 0) {
-	VALUE path;
+    if ((state = EXEC_TAG()) == TAG_NONE) {
 	long handle;
 	int found;
 
-	rb_set_safe_level_force(safe);
-	FilePathValue(fname);
 	rb_set_safe_level_force(0);
 
-	RUBY_DTRACE_HOOK(FIND_REQUIRE_ENTRY, StringValuePtr(fname));
-
-	path = rb_str_encode_ospath(fname);
+	RUBY_DTRACE_HOOK(FIND_REQUIRE_ENTRY, RSTRING_PTR(fname));
 	found = search_required(path, &path, safe);
+	RUBY_DTRACE_HOOK(FIND_REQUIRE_RETURN, RSTRING_PTR(fname));
 
-	RUBY_DTRACE_HOOK(FIND_REQUIRE_RETURN, StringValuePtr(fname));
 	if (found) {
 	    if (!path || !(ftptr = load_lock(RSTRING_PTR(path)))) {
 		result = 0;
@@ -994,7 +1003,7 @@ rb_require_internal(VALUE fname, int safe)
 
 		  case 's':
 		    handle = (long)rb_vm_call_cfunc(rb_vm_top_self(), load_ext,
-						    path, 0, path);
+						    path, VM_BLOCK_HANDLER_NONE, path);
 		    rb_ary_push(ruby_dln_librefs, LONG2NUM(handle));
 		    break;
 		}
@@ -1010,13 +1019,14 @@ rb_require_internal(VALUE fname, int safe)
 
     rb_set_safe_level_force(saved.safe);
     if (state) {
+	RB_GC_GUARD(fname);
 	/* never TAG_RETURN */
 	return state;
     }
 
-    th->errinfo = errinfo;
+    th->ec.errinfo = errinfo;
 
-    RUBY_DTRACE_HOOK(REQUIRE_RETURN, StringValuePtr(fname));
+    RUBY_DTRACE_HOOK(REQUIRE_RETURN, RSTRING_PTR(fname));
 
     return result;
 }
@@ -1037,6 +1047,7 @@ rb_require_safe(VALUE fname, int safe)
     int result = rb_require_internal(fname, safe);
 
     if (result > TAG_RETURN) {
+	if (result == TAG_RAISE) rb_exc_raise(rb_errinfo());
 	JUMP_TAG(result);
     }
     if (result < 0) {
@@ -1049,8 +1060,7 @@ rb_require_safe(VALUE fname, int safe)
 VALUE
 rb_require(const char *fname)
 {
-    VALUE fn = rb_str_new2(fname);
-    OBJ_FREEZE(fn);
+    VALUE fn = rb_str_new_cstr(fname);
     return rb_require_safe(fn, rb_safe_level());
 }
 
@@ -1099,7 +1109,7 @@ rb_mod_autoload(VALUE mod, VALUE sym, VALUE file)
     ID id = rb_to_id(sym);
 
     FilePathValue(file);
-    rb_autoload(mod, id, RSTRING_PTR(file));
+    rb_autoload_str(mod, id, file);
     return Qnil;
 }
 
